#!/opt/local/bin/perl -w

# $Id: sslgui.pl.in,v 1.7 2004/12/27 15:38:31 thivillon Exp $

use Gtk;
use IO::Pipe;
use strict;
use warnings;
use POSIX ":sys_wait_h";
use Getopt::Std;

my $true = 1;
my $false = 0;
my $file_ok = 0;

my $default_file = '.ssltunnelrc';
my $config = {};
my $tab = undef;
my $filename = undef;
my $output = undef;
my $hbox = undef;
my $input = undef;
my $textlen=0;
my $notebook = undef;
my $pid = 0;
my $nbcharmax = 10000;
my $initial_file;
my $handle_pipe = undef;

my %conf_control = ();

my %opts = ();

if (!&getopts('s', \%opts)) {
  print STDERR "Usage: gpppclient [-s] [config-file]\n";
  exit(1);
}

init Gtk;

my $window = new Gtk::Window( "toplevel" );
$window->signal_connect( "delete_event", \&file_quit);  

# set window attributes and show it
$window->border_width( 0 );
$window->set_default_size( 300, 300 ); 

my @menu_items = (
   {
      path => '/_File',
      type => '<Branch>',
   },
   {
      path => '/_Tunnel',
      type => '<Branch>',
   },
   {
      path => '/File/_Open',
      accelerator => '<control>O',
   },
   {
      path => '/File/_Close',
      accelerator => '<control>W',
   },
   {
      path => '/File/_Save',
      accelerator => '<control>S',
   },
   {
      path => '/File/Save _As...',
   },
   {
      path => '/File/_Quit',
      accelerator => '<control>Q',
   },
   {
      path => '/Tunnel/_Start',
      accelerator => '<control>K',
   },
   {
      path => '/Tunnel/S_top',
      accelerator => '<control>E',
   },
);

$notebook = new Gtk::Notebook(); 
$notebook->set_tab_pos( 'top' ); 
$notebook->show;

$tab = new Gtk::Table( 11, 4, $false ); 
$tab->set_row_spacings( 5 ); 
$tab->set_col_spacings( 5 ); 
$tab->show;

my $label = new Gtk::Label( 'Configuration' );
$label->show;
$notebook->append_page( $tab, $label ); 

my $item_f = get_main_menu( $window );
my $menu_bar = $item_f->get_widget( '<main>' );
my $open_item =  $item_f->get_widget( '/File/Open');
my $close_item =  $item_f->get_widget( '/File/Close');
my $save_item =  $item_f->get_widget( '/File/Save');
my $saveas_item =  $item_f->get_widget( '/File/Save As...');
my $quit_item =  $item_f->get_widget( '/File/Quit');
my $start_item = $item_f->get_widget( '/Tunnel/Start');
my $stop_item = $item_f->get_widget( '/Tunnel/Stop');

$open_item->signal_connect( 'activate', \&file_open  );
$close_item->signal_connect( 'activate', \&file_close, 1 );
$save_item->signal_connect( 'activate', \&file_save );
$saveas_item->signal_connect( 'activate', \&file_save_as );
$quit_item->signal_connect( 'activate', \&file_quit, 1 );
$start_item->signal_connect( 'activate', \&start_tunnel, 1 );
$stop_item->signal_connect( 'activate', \&stop_tunnel);

$menu_bar->show();

my $vbox = new Gtk::VBox( 0, 0 );
$vbox->pack_start( $menu_bar,  $false, $false, 0 );
$vbox->pack_start( $notebook,  $true, $true, 0 );
$vbox->show;

$output = new Gtk::Text( undef, undef );
my $vscrollbar = new Gtk::VScrollbar( $output->vadj );

$hbox = new Gtk::HBox ( $false , 0 );
$hbox->pack_start( $vscrollbar, $false, $false, 0 ); 
$hbox->pack_start( $output, $true, $true, 0 ); 
$hbox->show();
$vscrollbar->show(); 
$output->set_editable( $false );
$output->show();
$label = new Gtk::Label( 'Output' );
$label->show;
$notebook->append_page( $hbox, $label ); 

$window->add($vbox);

$window->show;

&create_controls();

if (!defined($ARGV[0]) || $ARGV[0] eq '') {
  $initial_file = $ENV{'HOME'} . '/' . $default_file;  
}
else {
  $initial_file = $ARGV[0];
}

$config = &load_config( $initial_file );
&fill_controls;
if (defined($config)) {
  $open_item->set_sensitive(0);
  $stop_item->set_sensitive(0);
  $filename = $initial_file;
  $window->set_title("[" . $config->{'peer'} . "] SSLTunnel");
}
else {
  $close_item->set_sensitive(0);
  $save_item->set_sensitive(0);
  $start_item->set_sensitive(0);
  $stop_item->set_sensitive(0);
  $window->set_title("SSLTunnel");
}


$window->show();

$SIG{CHLD} = \&reap_childs;

if ($opts{'s'} && $file_ok) {
  $start_item->activate;
}

# Gtk event loop
main Gtk;

exit( 0 );


sub file_quit {

   my $widget = shift;
   my $confirm = shift ;

   return 1 if (defined($pid) && $pid > 0);
   if ($confirm && &check_changes) {
     &dialog_changes( sub { &file_quit($widget, 0) } ); 
     return 1;
   }
   Gtk->exit( 0 );
   return 1;
}

sub load_config($) {

  my $filename=shift;

  my $config = {};
  $config->{'verbose'} = 1;
  $config->{'remotehost'} = '192.168.1.1';
  $config->{'port'} = 443;
  $config->{'localppp'} = '/usr/sbin/pppd';
  $config->{'bsdppp'} = 0;
  $config->{'peer'} = '';
  $config->{'ipparam'} = 'tunnel';
  $config->{'localproxyarp'} = 0;
  $config->{'localechoint'} = 10;
  $config->{'localechofail'} = 10;
  $config->{'useproxy'} = 0;
  $config->{'proxy'} = '';
  $config->{'proxyport'} = 8080;
  $config->{'proxyuser'} = '';
  $config->{'proxypass'} = '';
  $config->{'cacertfile'} = '';
  $config->{'keyfile'} = '';
  $config->{'certfile'} = '';
  $config->{'autoreconnect'} = 0;
  $config->{'daemon'} = 0;
  $config->{'logfile'} = '';

  return $config unless open CFG,"<$filename";
  my $line = '';
  while (defined($line=<CFG>)) {
    chomp $line;
    next if ($line =~ m#^\s*$#);
    next if ($line =~ m#\##);
    next unless ($line =~ /^(\S+)\s+(\S+)/);
    $config->{$1} = $2;
  }
  close CFG;
  $file_ok = 1;
  return $config;

}

sub file_open() {

 my $file_dialog = new Gtk::FileSelection( "File Selection" );

 # Connect the ok_button to file_ok_sel function
 $file_dialog->ok_button->signal_connect( "clicked",
					 \&file_ok_open,
					 $file_dialog );

 $file_dialog->cancel_button->signal_connect( "clicked",
                                              sub {  
                                                my $widget = shift; 
                                                my $dialog = shift; 
                                                $dialog->destroy();
                                              },
                                              $file_dialog,
                                            );

 $file_dialog->show();

}

sub file_ok_open {

   my $widget = shift;
   my $dialog = shift;
   my $local_filename = $dialog->get_filename(); 
   if (defined($local_filename) &&
       $local_filename ne '' &&
       -r $local_filename )  {
      if (defined($config = &load_config($local_filename))) {
         $open_item->set_sensitive(0);
         $close_item->set_sensitive(1);
         $save_item->set_sensitive(1);
         $saveas_item->set_sensitive(1);
         $start_item->set_sensitive(1);
         $stop_item->set_sensitive(0);
         $dialog->destroy();
         $filename = $local_filename;
         &fill_controls();
         $window->set_title("[" . $config->{'peer'} . "] SSLTunnel");
      }
   }
   return 0;
}

sub create_controls {

  my $verbose =  new Gtk::CheckButton( 'Verbose ouput' );
  $verbose->show();
  $tab->attach( $verbose, 0, 1, 0, 1, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'verbose'} = $verbose;

  my $autoreconnect =  new Gtk::CheckButton( 'Try to reconnect when link goes down' );
  $autoreconnect->show();
  $tab->attach( $autoreconnect, 1, 3, 0, 1, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'autoreconnect'} = $autoreconnect;

  my $text = new Gtk::Label('Destination');
  $text->set_justify( 'left' );
  $tab->attach( $text, 0, 1, 1, 2, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  $text->set_alignment(0.05, 0.5);
  my $remotehost = new Gtk::Entry( 63 );
  $remotehost->show();
  $tab->attach( $remotehost, 1, 2, 1, 2,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'remotehost'} = $remotehost;

  my $port = new Gtk::Entry( 8 );
  $port->show();
  $tab->attach( $port, 2, 3, 1, 2, [], [], 0, 0);
  $conf_control{'port'} = $port;

  $text = new Gtk::Label('PPP Program');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 2, 3, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  my $localppp = new Gtk::Entry( 200 );
  $localppp->show();
  $tab->attach( $localppp, 1, 2, 2, 3,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'localppp'} = $localppp;

  my $bsdppp =  new Gtk::CheckButton( 'Userland PPP (*BSD)' );
  $bsdppp->show();
  $tab->attach( $bsdppp, 2, 3, 2, 3,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'bsdppp'} = $bsdppp;

  $text = new Gtk::Label('PPP Peer/Ipparam');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 3, 4, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();

  my $peer = new Gtk::Entry( 20 );
  $peer->show();
  $tab->attach( $peer, 1, 2, 3, 4,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'peer'} = $peer;

  my $ipparam = new Gtk::Entry( 20 );
  $ipparam->show();
  $tab->attach( $ipparam, 2, 3, 3, 4,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'ipparam'} = $ipparam;


  my $proxy = new Gtk::Entry( 4*3+3 );
  $proxy->show();
  $tab->attach( $proxy, 1, 2, 4, 5,  ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $conf_control{'proxy'} = $proxy;

  my $proxyport = new Gtk::Entry( 5 );
  $tab->attach( $proxyport, 2, 3, 4, 5,  [], [], 0, 0);
  $proxyport->show();
  $conf_control{'proxyport'} = $proxyport;

  if (! $config->{'useproxy'}) {
    $proxy->set_sensitive(0);
    $proxyport->set_sensitive(0);
  }

  $text = new Gtk::Label('Proxy User/Password');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 5, 6, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  my $proxyuser = new Gtk::Entry( 4*3+3 );
  $proxyuser->show();
  $tab->attach( $proxyuser, 1, 2, 5, 6, ['fill'] ,['fill' ], 0, 0);
  $conf_control{'proxyuser'} = $proxyuser;

  my $proxypass = new Gtk::Entry( 12 );
  $tab->attach( $proxypass, 2, 3, 5, 6, ['expand'], ['expand'], 0, 0);
  $proxypass->set_visibility( $false ); 
  $proxypass->show();
  $conf_control{'proxypass'} = $proxypass;

  if (! $config->{'useproxy'}) {
    $proxyuser->set_sensitive(0);
    $proxypass->set_sensitive(0);
  }

  my $useproxy =  new Gtk::CheckButton( 'Use Proxy' );
  $useproxy->show();
  $tab->attach( $useproxy, 0, 1, 4, 5, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $useproxy->signal_connect( 'toggled', \&set_proxy,  [ $proxy, $proxyport, $proxyuser, $proxypass] );
  $conf_control{'useproxy'} = $useproxy;

  $text = new Gtk::Label('CA Certificate');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 6, 7, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  my $cacertfile = new Gtk::Entry( 128 );
  $cacertfile->show();
  $conf_control{'cacertfile'} = $cacertfile;
  $tab->attach( $cacertfile, 1, 2, 6, 7, ['fill', 'expand'] ,['fill', 'expand'], 0, 0);
  my $button = new Gtk::Button( '...' );
  $button->signal_connect( 'clicked', \&dialog_cert, 'cacertfile' );
  $tab->attach( $button, 2, 3, 6, 7, [] ,[], 0, 0);
  $button->show();

  $text = new Gtk::Label('Key file');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 8, 9, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  my $keyfile = new Gtk::Entry( 128 );
  $keyfile->show();
  $conf_control{'keyfile'} = $keyfile;
  $tab->attach( $keyfile, 1, 2, 8, 9, ['fill', 'expand'] ,['fill', 'expand'], 0, 0);
  $button = new Gtk::Button( '...' );
  $button->signal_connect( 'clicked', \&dialog_cert, 'keyfile' );
  $tab->attach( $button, 2, 3, 8, 9, [] ,[], 0, 0);
  $button->show();

  $text = new Gtk::Label('Certificate file');
  $text->set_justify( 'left' );
  $text->set_alignment(0.05, 0.5);
  $tab->attach( $text, 0, 1, 9, 10, ['fill', 'expand'], ['fill', 'expand'], 0, 0);
  $text->show();
  my $certfile = new Gtk::Entry( 128 );
  $certfile->show();
  $conf_control{'certfile'} = $certfile;
  $tab->attach( $certfile, 1, 2, 9, 10, ['fill', 'expand'] ,['fill', 'expand'], 0, 0);
  $button = new Gtk::Button( '...' );
  $button->signal_connect( 'clicked', \&dialog_cert, 'certfile' );
  $tab->attach( $button, 2, 3, 9, 10, [] ,[], 0, 0);
  $button->show();


}

sub fill_controls {

  $conf_control{'verbose'}->set_active( $config->{'verbose'} );
  $conf_control{'autoreconnect'}->set_active( $config->{'autoreconnect'} );

  $conf_control{'remotehost'}->set_text( $config->{'remotehost'} );
  $conf_control{'port'}->set_text( $config->{'port'} );
  $conf_control{'localppp'}->set_text( $config->{'localppp'} );
  $conf_control{'bsdppp'}->set_active( $config->{'bsdppp'} );

  $conf_control{'proxy'}->set_text( $config->{'proxy'} );
  $conf_control{'proxyport'}->set_text( $config->{'proxyport'} );

  $conf_control{'proxyuser'}->set_text( $config->{'proxyuser'} );
  $conf_control{'proxypass'}->set_text( $config->{'proxypass'} );

  $conf_control{'peer'}->set_text( $config->{'peer'} );
  $conf_control{'ipparam'}->set_text( $config->{'ipparam'} );

  if (! $config->{'useproxy'}) {
    $conf_control{'proxyuser'}->set_sensitive(0);
    $conf_control{'proxypass'}->set_sensitive(0);
  }

  $conf_control{'useproxy'}->set_active( $config->{'useproxy'} );
  $conf_control{'cacertfile'}->set_text( $config->{'cacertfile'} );
  $conf_control{'keyfile'}->set_text( $config->{'keyfile'} );
  $conf_control{'certfile'}->set_text( $config->{'certfile'} );

  $tab->show();
  $notebook->set_page( $notebook->page_num( $tab ));

}

sub set_proxy() {

  my $widget = shift;
  my $toactive = shift;

  my $curr = undef;
  $config->{'useproxy'} = $widget->active;
  foreach $curr (@$toactive) {
    $curr->set_sensitive($config->{'useproxy'});
  }

}

sub file_close() {

 my $widget = shift;
 my $confirm = shift;
  
 return 1 if (defined($pid) && $pid > 0);
 if ($confirm && &check_changes) {
    &dialog_changes( sub { &file_close($widget, 0) } ); 
    return 1;
 }

 $tab->hide();
 $open_item->set_sensitive(1);
 $close_item->set_sensitive(0);
 $save_item->set_sensitive(0);
 $saveas_item->set_sensitive(0);
 $start_item->set_sensitive(0);
 $stop_item->set_sensitive(0);
 $window->set_title("SSLTunnel");

}

sub file_save {

  if (!open CFG, ">$filename") {
    &dialog_alert("Unable to open $filename : $!");
    return 1;
  }
  
  $config->{'verbose'} = $conf_control{'verbose'}->active;
  $config->{'autoreconnect'} = $conf_control{'autoreconnect'}->active;
  $config->{'bsdppp'} = $conf_control{'bsdppp'}->active;
  $config->{'useproxy'} = $conf_control{'useproxy'}->active;

  foreach (qw/remotehost port localppp proxy proxyport 
              proxyuser proxypass cacertfile keyfile certfile
              peer ipparam/) {
     $config->{$_} = $conf_control{$_}->get_text();
  }

  my $key = undef;
  foreach $key (sort {$a cmp $b} keys(%$config)) {
     printf CFG "%s\t%s\n",$key,$config->{$key};
  }
  if (!close CFG) {
    &dialog_alert("Unable to close $filename : $!");
    return 1;
  }  

  load_config($filename);
  $window->set_title("[" . $config->{'peer'} . "] SSLTunnel");
  return 0;

}

sub file_save_as {

 my $file_dialog = new Gtk::FileSelection( "Save config as ...");

 $file_dialog->ok_button->signal_connect( "clicked",
                                         sub {
                                            my $widget = shift; 
                                            my $dialog = shift; 
                                            $filename = $dialog->get_filename();
                                            &file_save();
                                            $dialog->destroy();
                                         },
					 $file_dialog);

 $file_dialog->cancel_button->signal_connect( "clicked",
                                              sub {  
                                                my $widget = shift; 
                                                my $dialog = shift; 
                                                $dialog->destroy();
                                              },
                                              $file_dialog,
                                            );
 $file_dialog->set_filename($filename);
 $file_dialog->show();
 return 0;

}


sub dialog_cert {

 my $widget =  shift;
 my $whichfile = shift;

 my $file_dialog = new Gtk::FileSelection( "$whichfile Selection" );

 # Connect the ok_button to file_ok_sel function
 $file_dialog->ok_button->signal_connect( "clicked",
                                         \&choose_cert,
					 $file_dialog, $whichfile );

 $file_dialog->cancel_button->signal_connect( "clicked",
                                              sub {  
                                                my $widget = shift; 
                                                my $dialog = shift; 
                                                $dialog->destroy();
                                              },
                                              $file_dialog,
                                            );
 $file_dialog->set_filename($conf_control{$whichfile}->get_text());
 $file_dialog->show();

}

sub choose_cert {
  my $widget = shift;
  my $dialog = shift;
  my $whichfile = shift;

  $conf_control{$whichfile}->set_text($dialog->get_filename);
  $dialog->destroy;
}


sub dialog_changes() {
  
  my $message = "Settings have been changed, do you want to save ?";
  my $reffun = shift;

  my $dialog = new Gtk::Dialog(); 
  my $button_yes = new Gtk::Button( "Yes");
  my $button_no = new Gtk::Button( "No");
  my $button_cancel = new Gtk::Button( "Cancel");

  $button_yes->signal_connect( 'clicked',
      sub {
        my $widget = shift;
        my $dialog = shift;
        $dialog->destroy();
        if (!defined($filename) || $filename eq '') { &file_save_as() }
        else { &file_save() };
        &$reffun() ;
      },
      $dialog,
    );

  $button_no->signal_connect( 'clicked',
      sub {
        my $widget = shift;
        my $dialog = shift;
        $dialog->destroy();
        &$reffun();
      }  ,
      $dialog,
  );

  $button_cancel->signal_connect( 'clicked',
      sub {
        my $widget = shift;
        my $dialog = shift;
        $dialog->destroy();
      }  ,
      $dialog,
    );

  $dialog->action_area->pack_start( $button_yes, $true, $true, 0 );
  $dialog->action_area->pack_start( $button_no, $true, $true, 0 );
  $dialog->action_area->pack_start( $button_cancel, $true, $true, 0 );
  $button_yes->show();
  $button_no->show();
  $button_cancel->show();
 
  my $label = new Gtk::Label( $message );
  $dialog->vbox->pack_start( $label, $true, $true, 0 );
  $label->show();
  $dialog->set_modal( $true );
  $dialog->show();

}


sub dialog_alert() {
  
  my $message = shift;

  my $dialog = new Gtk::Dialog(); 
  my $button = new Gtk::Button( "Close");
  $button->signal_connect( 'clicked',
      sub {
        my $widget = shift;
        my $dialog = shift;
        $dialog->destroy();
      }  ,
      $dialog,
    );
  $dialog->action_area->pack_start( $button, $true, $true, 0 );
  $button->show();
  my $label = new Gtk::Label( $message );
  $dialog->vbox->pack_start( $label, $true, $true, 0 );
  $label->show();
  $dialog->set_modal( $true );
  $dialog->show();

}

sub start_tunnel {

   my $widget = shift;
   my $confirm = shift ;

   if ($confirm && &check_changes) {
     &dialog_changes( sub { &start_tunnel($widget, 0) } );
     return 1;
   }

   $start_item->set_sensitive(0);
   $stop_item->set_sensitive(1);
   $quit_item->set_sensitive(0);
   $notebook->set_page( $notebook->page_num( $hbox ));
  
   $pid = open(PPP,"-|");
   if ($pid) {
      $handle_pipe = \*PPP;
      $handle_pipe->blocking($false);
      $input = Gtk::Gdk->input_add(  fileno($handle_pipe) , 'read', 
         \&read_sslclient, $handle_pipe );
   }
   else {
      open STDERR, '>&STDOUT';
      exec('pppclient', '-d', '0', $filename) or Gtk->exit( 0 );
   }
}

sub read_sslclient {

  my $handle = shift;
  return unless (fileno($handle) == fileno($handle_pipe));
  if (defined(my $line = <$handle>)) {
     $textlen += length($line);
     $output->freeze();
     if ($textlen > $nbcharmax ) {
        $output->set_point( 0 );
        $output->forward_delete(length($line));
     }
     $output->set_point( $output->get_length()); 
     $output->thaw();
     $output->insert( undef , undef , undef , $line );
  }
  else {
    Gtk::Gdk->input_remove( $input ) if (defined($input));
    close $handle_pipe if (defined($handle_pipe));
    $start_item->set_sensitive(1);
    $stop_item->set_sensitive(0);
    $quit_item->set_sensitive(1);
    $output->set_point( $output->get_length()); 
    $output->insert( undef , undef , undef , "pppclient terminated\n" );
  }
}

sub stop_tunnel {
   if (defined($pid) && $pid > 0) {
     kill 15, $pid;
   }
}

sub check_changes {

  my $changed = 0;

  return 1 unless $file_ok;

  foreach (qw/verbose autoreconnect bsdppp useproxy/) {
     $changed |= ($config->{$_} != $conf_control{$_}->active);
  }
  foreach (qw/remotehost port localppp proxy proxyport 
              proxyuser proxypass cacertfile keyfile certfile
              peer ipparam/) {
     $changed |= ($config->{$_} ne $conf_control{$_}->get_text());
  }
  return $changed;

}

sub reap_childs {

  my $child;
  $pid = 0;
  while (($child = waitpid(-1,WNOHANG)) > 0) {
  }
  $SIG{CHLD} = \&reap_childs;
}

sub get_main_menu {
    my ( $window ) = @_;

    my $item_factory;
    my $accel_group;

    $accel_group = new Gtk::AccelGroup();

    $item_factory = new Gtk::ItemFactory( 'Gtk::MenuBar',
					  '<main>',
					  $accel_group );

    $item_factory->create_items( @menu_items );

    $window->add_accel_group( $accel_group );

    return ( $item_factory );
}

